package ch.psi.daq.queryrest.controller;

import java.io.IOException;
import java.util.Collection;
import java.util.LinkedHashSet;
import java.util.LinkedList;
import java.util.Map.Entry;
import java.util.Set;
import java.util.function.Function;
import java.util.stream.Stream;

import javax.annotation.Resource;
import javax.servlet.http.HttpServletResponse;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.http.MediaType;
import org.springframework.web.bind.annotation.RequestBody;
import org.springframework.web.bind.annotation.RequestMapping;
import org.springframework.web.bind.annotation.RequestMethod;
import org.springframework.web.bind.annotation.ResponseBody;
import org.springframework.web.bind.annotation.RestController;

import ch.psi.daq.cassandra.util.test.CassandraDataGen;
import ch.psi.daq.domain.DataEvent;
import ch.psi.daq.query.analyzer.QueryAnalyzer;
import ch.psi.daq.query.model.AbstractQuery;
import ch.psi.daq.query.model.Aggregation;
import ch.psi.daq.query.model.DBMode;
import ch.psi.daq.query.model.Query;
import ch.psi.daq.query.model.QueryField;
import ch.psi.daq.query.processor.QueryProcessor;
import ch.psi.daq.queryrest.config.QueryRestConfig;
import ch.psi.daq.queryrest.model.ChannelsRequest;
import ch.psi.daq.queryrest.response.ResponseStreamWriter;

@RestController
public class QueryRestController {

   private static final Logger LOGGER = LoggerFactory.getLogger(QueryRestController.class);

   public static final String CHANNELS = "channels";
   public static final String QUERY = "query";

   @Resource
   private ResponseStreamWriter responseStreamWriter;

   @Resource
   private QueryProcessor cassandraQueryProcessor;

   @Resource
   private QueryProcessor archiverApplianceQueryProcessor;

   @Resource
   private Function<Query, QueryAnalyzer> queryAnalizerFactory;

   @Resource(name = QueryRestConfig.BEAN_NAME_DEFAULT_RESPONSE_FIELDS)
   private Set<QueryField> defaultResponseFields;

   @Resource(name = QueryRestConfig.BEAN_NAME_DEFAULT_RESPONSE_AGGREGATIONS)
   private Set<Aggregation> defaultResponseAggregations;

   @RequestMapping(
         value = CHANNELS,
         method = {RequestMethod.GET, RequestMethod.POST},
         produces = {MediaType.APPLICATION_JSON_VALUE})
   public @ResponseBody Collection<String> getChannels(@RequestBody(required = false) ChannelsRequest request)
         throws Throwable {
      // in case not specified use default (e.g. GET) 
      if (request == null) {
         request = new ChannelsRequest();
      }

      try {
         return getQueryProcessor(request.getDbMode()).getChannels(request.getRegex());
      } catch (Throwable t) {
         LOGGER.error("Failed to query channel names.", t);
         throw t;
      }
   }

   @RequestMapping(
         value = QUERY,
         method = RequestMethod.POST,
         consumes = {MediaType.APPLICATION_JSON_VALUE},
         produces = {MediaType.APPLICATION_JSON_VALUE})
   public void executeQuery(@RequestBody AbstractQuery query, HttpServletResponse res) throws IOException {
      try {
         LOGGER.debug("Execute query '{}'", query.getClass().getSimpleName());

         QueryAnalyzer queryAnalizer = queryAnalizerFactory.apply(query);
         queryAnalizer.validate();

         extendQuery(query);

         // all the magic happens here
         Stream<Entry<String, Stream<? extends DataEvent>>> channelToDataEvents =
               getQueryProcessor(query.getDBMode()).process(queryAnalizer);

         // do post-process
         Stream<Entry<String, ?>> channelToData = queryAnalizer.postProcess(channelToDataEvents);

         // write the response back to the client using java 8 streams
         responseStreamWriter.respond(channelToData, query, res);
      } catch (Throwable t) {
         LOGGER.error("Failed execute query '{}'.", query, t);
         throw t;
      }
   }

   private QueryProcessor getQueryProcessor(DBMode dbMode) {
      if (DBMode.databuffer.equals(dbMode)) {
         return cassandraQueryProcessor;
      } else if (DBMode.archiverappliance.equals(dbMode)) {
         return archiverApplianceQueryProcessor;
      } else {
         LOGGER.error("Unknown DBMode '{}'!", dbMode);
         throw new IllegalArgumentException(String.format("Unknown DBMode '%s'", dbMode));
      }
   }

   private void extendQuery(AbstractQuery query) {
      if (query.getFields() == null || query.getFields().isEmpty()) {
         query.setFields(new LinkedHashSet<>(defaultResponseFields));
      }
      if (query.getAggregations() == null || query.getAggregations().isEmpty()) {
         query.setAggregations(new LinkedList<>(defaultResponseAggregations));
      }
   }

   // ==========================================================================================
   // TODO: This is simply for initial / rudimentary testing - remove once further evolved

   @Resource
   private CassandraDataGen cassandraDataGen;

   @RequestMapping(value = "/write")
   public void writeDummyEntry() {
      cassandraDataGen.writeData(3, 0, 100, "channel1", "channel2");
   }
}
